import { App, segment, Session, sleep } from 'koishi';
import onebot, { WsClient, OneBotBot, CQCode } from '@koishijs/plugin-adapter-onebot';

import { parseCmd, query, view, resendLast } from './command';
import { getLogger } from './loggers';

const logger = getLogger('qqbot');

interface IQQProps {
  access_token: string;
  host: string;
  port: number;
  bot_id: number;
  list(chat: IChat, args: string[], replyfn: (msg: string) => any): void;
  sub(chat: IChat, args: string[], replyfn: (msg: string) => any): void;
  unsub(chat: IChat, args: string[], replyfn: (msg: string) => any): void;
  unsubAll(chat: IChat, args: string[], replyfn: (msg: string) => any): void;
}

export const Message = {
  Image: segment.image,
  Video: segment.video,
  Voice: segment.audio,
  separateAttachment: (msg: string) => {
    const attachments: string[] = [];
    const message = msg.replace(/<(video|record) .*?\/>/g, code => {
      attachments.push(code);
      return '';
    });
    return {message, attachments};
  },
  parseCQCode: (cqStr: string) => CQCode.parse(cqStr).map(seg => {
    if (typeof seg.attrs.file === 'string') {
      seg.attrs.url = seg.attrs.file;
      delete seg.attrs.file;
    }
    return seg;
  }).join(''),
  toCQCode: (segStr: string) => segment.parse(segStr).map(seg => {
    if (typeof seg.attrs.url === 'string') {
      seg.attrs.file = seg.attrs.url;
      delete seg.attrs.url;
    }
    return CQCode(seg.type, seg.attrs);
  }).join(''),
};

export default class {

  private botInfo: IQQProps;
  private app: App;
  public bot: OneBotBot;

  private messageQueues: {[key: string]: (() => Promise<void>)[]} = {};
  private tempSenders: {[key: number | string]: number} = {};

  private get config(): onebot.Config & WsClient.Config {
    return {
      protocol: 'ws',
      endpoint: `ws://${this.botInfo.host}:${this.botInfo.port}`,
      selfId: this.botInfo.bot_id.toString(),
      token: this.botInfo.access_token,
    };
  };

  private next = (type: ChatType, id: string) => {
    const queue = this.messageQueues[`${type}:${id}`];
    if (queue && queue.length) {
      queue[0]().then(() => {
        queue.shift();
        if (!queue.length) delete this.messageQueues[`${type}:${id}`];
        else this.next(type, id);
      });
    }
  };

  private enqueue = (type: ChatType, id: string, resolver: () => Promise<void>) => {
    let wasEmpty = false;
    const queue = this.messageQueues[`${type}:${id}`] ||= (() => { wasEmpty = true; return []; })();
    queue.push(() => sleep(200).then(resolver));
    logger.debug(`no. of message currently queued for ${type}:${id}: ${queue.length}`);
    if (wasEmpty) this.next(type, id);
  };

  private getSender = ({chatID, chatType}: IChat) =>
    (msg: string) => (
      chatType === 'guild' ? this.sendToChannel :
      chatType === 'group' ? this.sendToGroup :
      this.sendToUser
    )(chatID.toString(), msg);

  private getChat = async (session: Session): Promise<IChat> => {
    if (session.type !== 'message') return null;
    if (session.subsubtype === 'guild') {
      return {
        chatID: `${session.onebot.guild_id}:${session.onebot.channel_id}`,
        chatType: 'guild',
      };
    }
    switch (session.subtype) {
      case 'private':
        const {group_id: groupId} = session.onebot.sender as any;
        if (groupId) { // temp message
          const friendList = await session.bot.getFriendList();
          if (!friendList.some(friendItem => friendItem.userId === session.userId)) {
            this.tempSenders[session.userId] = groupId;
            return {
              chatID: {
                qq: Number(session.userId),
                group: Number(groupId),
                toString: () => session.userId,
              },
              chatType: 'temp',
            };
          }
        }
        return { // already befriended
          chatID: Number(session.userId),
          chatType: 'private',
        };
      case 'group':
        return {
          chatID: Number(session.guildId),
          chatType: 'group',
        };
    }
  };

  private sendToChannel = (guildChannel: string, message: string) => new Promise<string>((resolve, reject) => {
    const [guildID, channelID] = guildChannel.split(':');
    this.enqueue('guild', guildChannel, () => this.bot.guildBot.sendMessage(channelID, message, guildID)
      .then(([response]) => resolve(response)).catch(reject));
  });

  private sendToGroup = (groupID: string, message: string) => new Promise<string>((resolve, reject) => {
    this.enqueue('group', groupID, () => this.bot.sendMessage(groupID, message)
      .then(([response]) => resolve(response)).catch(reject));
  });

  private sendToUser = (userID: string, message: string) => new Promise<string>((resolve, reject) => {
    this.enqueue('private', userID, () => this.bot.sendPrivateMessage(userID, message)
    .then(([response]) => resolve(response)).catch(reject));
  });

  public sendTo = (subscriber: IChat, messageChain: string, noErrors = false) => Promise.all(
    (splitted => [splitted.message, ...splitted.attachments])(
      Message.separateAttachment(messageChain)
    ).map(this.getSender(subscriber)))
    .then(response => {
      if (response === undefined) return;
      logger.info(`pushing data to ${JSON.stringify(subscriber.chatID)} was successful, response: ${response}`);
    })
    .catch(reason => {
      logger.error(`error pushing data to ${JSON.stringify(subscriber.chatID)}, reason: ${reason}`);
      if (!noErrors) throw reason instanceof Error ? reason : Error(reason);
    });

  private initBot = () => {
    this.app = new App();
    this.app.plugin(onebot, this.config);

    this.app.on('friend-request', async session => {
      const userString = `${session.username}(${session.userId})`;
      logger.debug(`detected new friend request event: ${userString}`);
      try {
        const isTemp = session.username in this.tempSenders;
        const {guildId, guildName} = isTemp ?
          await session.bot.getGuild(this.tempSenders[session.userId].toString()) :
            (await session.bot.getGuildList()).find(async ({guildId}) => {
              try {
                return await this.bot.getGuildMember(guildId, session.userId);
              } catch {}
            });
        try {
          const groupString = `${guildName}(${guildId})`;
          await session.bot.handleFriendRequest(session.messageId, true);
          logger.info(`accepted friend request from ${userString} (${
            isTemp ? 'from' : 'found in'
          } group ${groupString})`);
        } catch (error) {
          logger.error(`error accepting friend request from ${userString}, error: ${error}`);
        }
      } catch {
        logger.warn(`received friend request from ${userString} (stranger)`);
        logger.warn('please manually accept this friend request');
      }
    });

    this.app.on('guild-request', async session => {
      const userString = `${session.username}(${session.userId})`;
      const groupString = `${session.guildName}(${session.guildId})`;
      logger.debug(`detected group invitation event: ${groupString}}`);
      const friendList = await session.bot.getFriendList();
        if (friendList.some(friendItem => friendItem.userId = session.userId)) {
          try {
            session.bot.handleGuildRequest(session.messageId, true);
            logger.info(`accepted group invitation from ${userString} (friend)`);
          } catch (error) {
            logger.error(`error accepting group invitation from ${userString}, error: ${error}`);
          }
        }
        logger.warn(`received group invitation from ${userString} (stranger)`);
        logger.warn('please manually accept this group invitation');
    });

    this.app.middleware(async (session: Session) => {
      const chat = await this.getChat(session);
      let userString = `${session.username}(${session.userId})`;
      if (chat.chatType === 'temp') {
        const group = await session.bot.getGuild(chat.chatID.group.toString());
        userString += ` (from group ${group.guildName}(${group.guildId}))`;
      }
      const cmdObj = parseCmd(session.content);
      const reply = (msg: string) => this.getSender(chat)(msg).catch(error => {
        if (chat.chatType === 'temp') {
          return logger.info(`ignored error while replying to ${userString}`);
        }
        logger.error(`error replying to message from ${userString}, error: ${error}`);
      });
      switch (cmdObj.cmd) {
        case 'twi_view':
          view(chat, cmdObj.args, reply);
          break;
        case 'twi_resendlast':
          resendLast(chat, cmdObj.args, reply);
          break;
        case 'twi_query':
          query(chat, cmdObj.args, reply);
          break;
        case 'twi_sub':
          this.botInfo.sub(chat, cmdObj.args, reply);
          break;
        case 'twi_unsub':
          this.botInfo.unsub(chat, cmdObj.args, reply);
          break;
        case 'twi_unsuball':
          this.botInfo.unsubAll(chat, cmdObj.args, reply);
          break;
        case 'twi_listsub':
          this.botInfo.list(chat, cmdObj.args, reply);
          break;
        case 'help':
          if (cmdObj.args[0] === 'twi') {
            reply(`推特搬运机器人：
/twi_listsub - 查询当前聊天中的推文订阅
/twi_sub〈链接|用户名〉- 订阅 Twitter 推文搬运
/twi_unsub〈链接|用户名〉- 退订 Twitter 推文搬运
/twi_view〈链接|表达式〉[{force|refresh}={on|off}] - 查看推文（可选强制重新载入）
/twi_resendlast〈用户名〉- 强制重发该用户最后一条推文
/twi_query〈链接|用户名〉[参数列表...] - 查询时间线（详见 /help twi_query）\
${chat.chatType === 'temp' ?
    '\n（当前游客模式下无法使用订阅功能，请先添加本账号为好友。）' : ''
}`);
          } else if (cmdObj.args[0] === 'twi_query') {
            reply(`查询时间线中的推文：
/twi_query〈链接|用户名〉[〈参数 1〉=〈值 1〉〈参数 2〉=〈值 2〉...]

参数列表（方框内全部为可选，留空则为默认）：
    count：查询数量上限（类型：非零整数，最大值正负 50）[默认值：10]
    since：查询起始点（类型：正整数或日期）[默认值：（空，无限过去）]
    until：查询结束点（类型：正整数或日期）[默认值：（空，当前时刻）]
    noreps 忽略回复推文（类型：on/off）[默认值：on（是）]
    norts：忽略原生转推（类型：on/off）[默认值：off（否）]`)
              .then(() => reply(`\
起始点和结束点为正整数时取推特推文编号作为比较基准，否则会尝试作为日期读取。
推荐的日期格式：2012-12-22 12:22 UTC+2 （日期和时间均为可选，可分别添加）
count 为正时，从新向旧查询；为负时，从旧向新查询
count 与 since/until 并用时，取二者中实际查询结果较少者
例子：/twi_query RiccaTachibana count=5 since="2019-12-30\
 UTC+9" until="2020-01-06 UTC+8" norts=on
    从起始时间点（含）到结束时间点（不含）从新到旧获取最多 5 条推文，\
其中不包含原生转推（实际上用户只发了 1 条）`)
              );
          }
      }
    }, true);
  };

  private listen = async (logMsg = 'connecting to bot provider...'): Promise<void> => {
    logger.warn(logMsg);
    try {
      await this.app.start();
    } catch (err) {
      logger.error(`error connecting to bot provider at ${this.config.endpoint}, will retry in 2.5s...`);
      await sleep(2500);
      await this.listen('retry connecting...');
    }
  };

  public connect = async () => {
    this.initBot();
    await this.listen();
    this.bot = this.app.bots.find(bot => bot.selfId === this.config.selfId) as OneBotBot;
  };

  constructor(opt: IQQProps) {
    logger.warn(`Initialized koishi on ${opt.host}:${opt.port} with access_token ${opt.access_token}`);
    this.botInfo = opt;
  }
}
